import typing as t
from datetime import datetime
from json import dumps, loads
from pathlib import Path

from .._globals import IGNORE_LOCALS
from .._log_policy import LogPolicy

if t.TYPE_CHECKING:
    from .._traffic import Traffic


class JSONStore:
    filename: str
    location: t.Optional[t.Union[str, Path]]

    filepath: Path
    log_policy: LogPolicy

    _traffic_instance = None

    def __repr__(self) -> str:
        return f"<JSONStore filename={self.filepath} log_policy={self.log_policy}>"

    def __init__(
        self,
        filename: str = "traffic.json",
        location: t.Optional[t.Union[str, Path]] = None,
        log_policy: LogPolicy = None,
    ) -> None:
        """
        Create a new JSONStore instance.

        :param filename: the name of the JSON file to store the traffic data in
        :param location: the location of the JSON file
        :param log_policy: the log policy to use (defaults to log everything if not provided)
        """
        self.filename = filename
        self.location = location

        if log_policy is None:
            from .._log_policy import LogPolicy

            self.log_policy = LogPolicy()

        else:
            self.log_policy = log_policy

    def setup(self, traffic_instance: "Traffic") -> None:
        """
        Set up the JSONStore instance.

        :param traffic_instance:
        :return:
        """
        # set filepath to instance folder if location is None
        if self.location is None:
            self.filepath = traffic_instance.app_instance_folder / self.filename
        else:
            # This expects an absolute path
            if isinstance(self.location, str):
                self.filepath = Path(self.location) / self.filename

            # This expects a Path object
            if isinstance(self.location, Path):
                self.filepath = self.location / self.filename

        # create the file if it doesn't exist
        if not self.filepath.exists():
            # create the parent directory if it doesn't exist
            if not self.filepath.parent.exists():
                self.filepath.parent.mkdir(parents=True)

            # file is created here
            self.filepath.touch()

    def log(
        self,
        request_date: t.Optional[datetime] = None,
        request_method: t.Optional[str] = None,
        request_host_url: t.Optional[str] = None,
        request_path: t.Optional[str] = None,
        request_endpoint: t.Optional[str] = None,
        request_remote_address: t.Optional[str] = None,
        request_referrer: t.Optional[str] = None,
        request_user_agent: t.Optional[str] = None,
        request_browser: t.Optional[str] = None,
        request_platform: t.Optional[str] = None,
        response_time: t.Optional[int] = None,
        response_size: t.Optional[str] = None,
        response_status_code: t.Optional[int] = None,
        response_exception: t.Optional[str] = None,
        response_mimetype: t.Optional[str] = None,
    ) -> None:
        """
        Log the traffic data.

        :param request_date: the date and time of the request
        :param request_method: the HTTP method of the request
        :param request_host_url: the host URL of the request
        :param request_path: the path of the request
        :param request_endpoint: the matched endpoint of the request
        :param request_remote_address: the remote address of the request
        :param request_referrer: the referrer of the request
        :param request_user_agent: the user agent of the request
        :param request_browser: the browser of the request (if able to be determined)
        :param request_platform: the platform of the request (if able to be determined)
        :param response_time: the amount of time in milliseconds it took to respond to the request
        :param response_size: the size of the response
        :param response_status_code: the status code of the response
        :param response_exception: the exception that occurred (if any)
        :param response_mimetype: the mimetype of the response
        :return:
        """
        data = {}

        for attr, attr_val in self.log_policy.__dict__.items():
            if attr in IGNORE_LOCALS:
                continue

            if attr_val:
                if isinstance(locals()[attr], datetime):
                    data[attr] = locals()[attr].isoformat()
                    continue

                data[attr] = locals()[attr]

            else:
                continue

        file_contents = self.filepath.read_text()

        if file_contents == "":
            jsond = []
        else:
            jsond = loads(file_contents)

        jsond.append(data)
        self.filepath.write_text(dumps(jsond, indent=0))

    def read(self) -> t.List[t.Dict[str, t.Any]]:
        """
        Read the JSON file and return the contents.

        :return: the contents of the JSON file
        """
        file_contents = self.filepath.read_text()

        if file_contents == "":
            return []

        logs = loads(file_contents)

        if isinstance(logs, list):
            return list(reversed(logs))

        return []



import timeit
import typing as t
from datetime import datetime
from pathlib import Path

from flask import Flask, g, request
from flask import Response

from .stores._protocols import StoreProtocol


class Traffic:
    """
    The Traffic class registers the extension with the Flask app and
    sets up any passed in stores.
    """

    app: Flask
    app_instance_folder: t.Optional[Path]

    stores: t.Union[StoreProtocol, t.List[StoreProtocol]]

    def __repr__(self) -> str:
        return f"<Traffic stores={self.stores}>"

    def __init__(
        self,
        app: t.Optional[Flask] = None,
        stores: t.Optional[t.Union[StoreProtocol, t.List[StoreProtocol]]] = None,
    ) -> None:
        """
        Initialize the flask-traffic extension.

        :param app: the flask app to attach the extension to
        :param stores: a single store or stores to log traffic data to
        """
        if app is not None:
            if stores is None:
                raise ImportError("No stores were passed in.")
            self.init_app(app, stores)

    def init_app(
        self,
        app: Flask,
        stores: t.Union[t.Union[StoreProtocol, t.List[StoreProtocol]]],
    ) -> None:
        """
        Initialize the flask-traffic extension.

        :param app: the flask app to attach the extension to
        :param stores: a single store or stores to log traffic data to
        """
        if app is None:
            raise ImportError(
                "No app was passed in, do traffic = Traffic(flaskapp) or traffic.init_app(app)"
            )
        if not isinstance(app, Flask):
            raise TypeError("The app that was passed in is not an instance of Flask")

        self.app = app
        self.app.extensions["traffic"] = self
        self.app_instance_folder = Path(app.instance_path)

        if isinstance(stores, list):
            self.stores = stores
        else:
            self.stores = [stores]

        self._setup_stores()
        self._setup_request_watcher()

    def _setup_stores(self) -> None:
        """
        Private method.
        Set up the stores that were passed in.
        :return:
        """
        for store in self.stores:
            store.setup(self)

    def _setup_request_watcher(self) -> None:
        """
        Private method.
        Set up before, after and teardown functions.
        :return:
        """

        @self.app.before_request
        def traffic_before_request():
            g.traffic_timer = timeit.default_timer()

        @self.app.after_request
        def traffic_after_request(response) -> Response:
            for store in self.stores:
                # If on_endpoints is set, only log requests to those endpoints
                if store.log_policy.on_endpoints:
                    if not request.endpoint in store.log_policy.on_endpoints:
                        continue

                # If skip_endpoints is set, do not log requests to those endpoints
                if store.log_policy.skip_endpoints:
                    if request.endpoint in store.log_policy.skip_endpoints:
                        continue

                # If on_status_codes is set, only log requests with those status codes
                if store.log_policy.on_status_codes:
                    if not response.status_code in store.log_policy.on_status_codes:
                        continue

                # If skip_status_codes is set, do not log requests with those status codes
                if store.log_policy.skip_status_codes:
                    if response.status_code in store.log_policy.skip_status_codes:
                        continue

                # If only_on_exception is set, skip the log here. See teardown_request below
                if not store.log_policy.only_on_exception:
                    store.log(
                        request_date=datetime.now(),
                        request_method=request.method,
                        request_host_url=request.host_url,
                        request_path=request.path,
                        request_endpoint=request.endpoint,
                        request_user_agent=request.user_agent.string,
                        request_remote_address=request.remote_addr,
                        request_referrer=request.referrer,
                        request_browser=request.user_agent.browser,
                        request_platform=request.user_agent.platform,
                        response_time=int(
                            (timeit.default_timer() - g.traffic_timer) * 1000
                        ),
                        response_size=response.content_length,
                        response_status_code=response.status_code,
                        response_exception=None,
                        response_mimetype=response.mimetype,
                    )

            return response

        @self.app.teardown_request
        def traffic_teardown_request(exception: t.Any) -> None:
            if exception:
                # attempt to pull the __repr__ of the exception
                try:
                    message = exception.__repr__()
                except AttributeError:
                    message = str(exception)

                for store in self.stores:
                    # If skip_on_exception is set, skip the log here
                    if not store.log_policy.skip_on_exception:
                        store.log(
                            request_date=datetime.now(),
                            request_method=request.method,
                            request_host_url=request.host_url,
                            request_path=request.path,
                            request_endpoint=request.endpoint,
                            request_user_agent=request.user_agent.string,
                            request_remote_address=request.remote_addr,
                            request_referrer=request.referrer,
                            request_browser=request.user_agent.browser,
                            request_platform=request.user_agent.platform,
                            response_time=int(
                                (timeit.default_timer() - g.traffic_timer) * 1000
                            ),
                            response_size=None,
                            response_status_code=500,
                            response_exception=message,
                            response_mimetype=None,
                        )



from flask import Flask, render_template
from flask_sqlalchemy import SQLAlchemy

from flask_traffic import Traffic, LogPolicy
from flask_traffic.stores import JSONStore, CSVStore, SQLStore, SQLORMStore, \
    SQLORMModelMixin, RedisStore

# create an instance of the flask_sqlalchemy extension
db = SQLAlchemy()

traffic = Traffic()


# this model is for the SQLORMStore set below
class ModelModel(db.Model, SQLORMModelMixin):
    pass


# create a log policy to pass the stores.
# This is used to disable all, then enable what you want.
log_policy = LogPolicy(
    skip_endpoints=("static",),
    skip_on_exception=True,
).set_from_false(
    request_date=True,
    request_browser=True,
    response_time=True,
    response_size=True,
    response_status_code=True,
    request_path=True,
)

only_on_exception = LogPolicy(
    only_on_exception=True,
).set_from_false(
    request_date=True,
    request_host_url=True,
    request_path=True,
    response_exception=True,
    response_time=True,
)

# create a csv file store
csv_store = CSVStore(log_policy=log_policy)

# create a sqlite store
sql_store = SQLStore(log_policy=log_policy)

# create a JSON store
json_store = JSONStore(log_policy=log_policy)

# create an ORM store and pass the above model
sqlorm_store = SQLORMStore(model=ModelModel)

# create a JSON store
json_exception_store = JSONStore(
    filename="exception.json", log_policy=only_on_exception)

# This example is configure to connect to a docker instance of redis
redis_store = RedisStore(
    redis_host="localhost",
    event_name="traffic",
    redis_port=8001,
    log_policy=log_policy,
)


# create a custom store and override the read method
class MyStore(JSONStore):
    def read(self):
        return "This is a custom store"


store_read_override = MyStore(filename="custom.json", log_policy=log_policy)


def create_app() -> Flask:
    app = Flask(__name__)

    app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite:///traffic_orm.sqlite"
    db.init_app(app)

    with app.app_context():
        db.create_all()

    # place the traffic extension below the db.init_app(app) line,
    # this will pick up th db.session automatically from db.init_app(app)
    # traffic.init_app(app, stores=[json_store, csv_store, sqlite_store, orm_store])

    traffic.init_app(app, stores=[
        csv_store,
        json_store,
        sql_store,
        sqlorm_store,
        redis_store,
        json_exception_store,
        store_read_override
    ])

    # You can add multiple stores at once, and they will all log data
    # based on the log policy

    @app.route("/")
    def index():
        return render_template("index.html")

    # This will create an exemption, and be stored in the json_exception_store
    @app.route("/exception")
    def exception():
        return render_template("exception.html")

    #
    # Reading will always be one behind, as the log is written after the request

    @app.route("/read-csv")
    def read_csv():
        return csv_store.read()

    @app.route("/read-json")
    def read_json():
        return json_store.read()

    @app.route("/read-sql")
    def read_sql():
        return sql_store.read()

    @app.route("/read-orm")
    def read_orm():
        return sqlorm_store.read()

    @app.route("/read-redis")
    def read_redis():
        return redis_store.read()

    @app.route("/read-custom")
    def read_custom():
        return store_read_override.read()

    @app.route("/read-exceptions")
    def read_exceptions():
        return json_exception_store.read()

    return app
